/* $Id: init.c,v 1.1 2000/12/10 11:31:36 marcus Exp $
******************************************************************************

   GGIGL extension initialization

   Copyright (C) 1998      Andreas Beck		[becka@ggi-project.org]
   Copyright (C) 1999-2000 Marcus Sundberg	[marcus@ggi-project.org]
  
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <stdio.h>
#include <string.h>

#include "config.h"
#include <ggi/internal/ggigl.h>
#include <ggi/internal/ggigl_debug.h>
#include <ggi/internal/gg_replace.h>	/* for snprintf() */


/* Exported variables
 */

uint32_t _ggiglDebug = 0;


/* Static variables 
 */
static int 	 _ggiglLibIsUp       = 0;
static void	*_ggiglConfigHandle  = NULL;
static char      _ggiglconfstub[512] = GGIGLCONFDIR;
static char     *_ggiglconfdir       = _ggiglconfstub+GGIGLTAGLEN;

/* Extension ID. Defaulting to -1 should make segfault on abuse more likely...
   This is exported so that sublibs can use it.
 */
ggi_extid _ggiglID = -1;

#define SUBLIB_PREFIX	"GGIGLdl_"


/*
 * Returns the directory where global config files are kept
 */

const char *ggiglGetConfDir(void)
{
#if defined(__WIN32__) && !defined(__CYGWIN__)
	/* On Win32 we allow overriding of the compiled in path. */
	const char *envdir = getenv("GGI_CONFDIR");
	if (envdir) return envdir;
#endif
	return _ggiglconfdir;
}

/* Dummy function which returns GGI_ENOFUNC
   We use this to reset the function pointers */
static int dummyfunc(void)
{
	return GGI_ENOFUNC;
}

/* Reset all function pointers to dummyfunc */
static void clearfuncs(ggiglpriv *priv)
{
	priv->swapbuffers = (void *) dummyfunc; 
}


/* This function is called by LibGGI when the API-list changes */
static int changed(ggi_visual_t vis, int whatchanged)
{
	DPRINT("changed(%p, %i) called for ggigl extension\n",
		vis, whatchanged);

	switch (whatchanged) {
	case GGI_CHG_APILIST: {
		int i;
		ggi_dlhandle *lib;
		char api[GGI_MAX_APILEN], args[GGI_MAX_APILEN];

		clearfuncs(GGIGLEXT_PRIV(vis));

		for (i=0; ggiGetAPI(vis, i, api, args) == 0; i++) {
			ggstrlcat(api,"-ggigl",sizeof(api));
			DPRINT_LIBS("Trying #%d: %s(%s)\n", i, api, args);

			lib = ggiExtensionLoadDL(vis, _ggiglConfigHandle,
					api, args, NULL, SUBLIB_PREFIX);
			if (lib != NULL) break;
		}
		}
		break;
	}
                                
	return 0;
}


void _ggiglInitBuiltins(void);
void _ggiglExitBuiltins(void);



/* Initialize the extension
 */
int ggiglInit(void)
{
	int err;
	const char *confdir;
	char *conffile;
	char *str;

	_ggiglLibIsUp++;
	if (_ggiglLibIsUp > 1) return 0; /* Initialize only at first call. */


	str = getenv("GGIGL_DEBUGSYNC");
	if (str != NULL) {
		_ggiglDebug |= DEBUG_SYNC;
	}

	str = getenv("GGIGL_DEBUG");
	if (str != NULL) {
		_ggiglDebug |= atoi(str) & DEBUG_ALL;
		DPRINT_CORE("%s Debugging=%d\n",
			    DEBUG_ISSYNC ? "sync" : "async",
			    _ggiglDebug);
	}

	confdir = ggiglGetConfDir();
	conffile = malloc(strlen(confdir) + 1 + strlen(GGIGLCONFFILE)+1);
	if (!conffile) {
		fprintf(stderr, "libggigl: unable to allocate memory for config filename.\n");
	} else {
		snprintf(conffile, strlen(confdir) + strlen(GGIGLCONFFILE) + 2,
			"%s/%s", confdir, GGIGLCONFFILE);

		err = ggLoadConfig(conffile, &_ggiglConfigHandle);
		if (err != GGI_OK) {
			fprintf(stderr,"LibGGIGL: couldn't open %s\n",
				conffile);
			_ggiglLibIsUp--;
			return err;
		}
		free(conffile);
	}

	_ggiglID = ggiExtensionRegister("GGIGL", sizeof(ggiglpriv), changed);

	if (_ggiglID < 0) {
		fprintf(stderr,
			"LibGGIGL: failed to register as extension.\n");
		_ggiglLibIsUp--;
		ggFreeConfig(_ggiglConfigHandle);
		return _ggiglID;
	}

	_ggiglInitBuiltins();

	return 0;
}


/* Deinitialize the extension
 */
int ggiglExit(void)
{
	int rc;

	if (!_ggiglLibIsUp) return -1;

	if (_ggiglLibIsUp > 1) {
		/* Exit only at last call */
		_ggiglLibIsUp--;
		return 0;
	}

	_ggiglExitBuiltins();

	rc = ggiExtensionUnregister(_ggiglID);
	ggFreeConfig(_ggiglConfigHandle);

	_ggiglLibIsUp = 0;

	return rc;
}


/* Attach the extension to a visual
 */
int
ggiglAttach(ggi_visual_t vis, int *attribs)
{
	int rc;

	rc = ggiExtensionAttach(vis, _ggiglID);
	if (rc == 0) {
		/* We are actually creating the primary instance. */
		memset(GGIGLEXT_PRIV(vis), 0, sizeof(ggiglpriv));
		GGIGLEXT_PRIV(vis)->priv = NULL;

		/* Now fake an "API change" so the right libs get loaded */
		changed(vis, GGI_CHG_APILIST);
	}

	/* Set attribs */
	ggiglSetAttribs(vis, attribs);

	return rc;
}
